@@ -86,7 +86,8 @@ def cachier(
     mongetter=None,
     cache_dir=None,
     hash_params=None,
-    wait_for_calc_timeout=0
+    wait_for_calc_timeout=0,
+    separate_files=False,
 ):
     """A persistent, stale-free memoization decorator.
 
@@ -131,32 +132,33 @@ def cachier(
         automatically hashable by Python.
     wait_for_calc_timeout: int, optional, for MongoDB only
         The maximum time to wait for an ongoing calculation. When a
-        process started to calculate the value setting being_calculated to 
-        True, any process trying to read the same entry will wait a maximum of 
+        process started to calculate the value setting being_calculated to
+        True, any process trying to read the same entry will wait a maximum of
         seconds specified in this parameter. 0 means wait forever.
         Once the timeout expires the calculation will be triggered.
+    separate_files: bool, default False, for Pickle cores only
+        Instead of a single cache file per-function, each function's cache is
+        split between several files, one for each argument set. This can help
+        if you per-function cache files become too large.
     """
-    # print('Inside the wrapper maker')
-    # print('mongetter={}'.format(mongetter))
-    # print('stale_after={}'.format(stale_after))
-    # print('next_time={}'.format(next_time))
-
     # The default is calculated dynamically to maintain previous behavior
     # to default to pickle unless the ``mongetter`` argument is given.
     if backend is None:
         backend = 'pickle' if mongetter is None else 'mongo'
-
     if backend == 'pickle':
         core = _PickleCore(  # pylint: disable=R0204
             stale_after=stale_after,
             next_time=next_time,
             reload=pickle_reload,
             cache_dir=cache_dir,
+            separate_files=separate_files
         )
     elif backend == 'mongo':
         if mongetter is None:
-            raise MissingMongetter('must specify ``mongetter`` when using the mongo core')
-        core = _MongoCore(mongetter, stale_after, next_time, wait_for_calc_timeout)
+            raise MissingMongetter(
+                'must specify ``mongetter`` when using the mongo core')
+        core = _MongoCore(
+            mongetter, stale_after, next_time, wait_for_calc_timeout)
     elif backend == 'memory':
         core = _MemoryCore(stale_after=stale_after, next_time=next_time)
     elif backend == 'redis':
