@@ -6,7 +6,7 @@
 # Licensed under the MIT license:
 # http://www.opensource.org/licenses/MIT-license
 # Copyright (c) 2016, Shay Palachy <shaypal5@gmail.com>
-
+import hashlib
 import os
 import pickle  # for local caching
 from datetime import datetime
@@ -83,7 +83,8 @@ class _PickleCore(_BaseCore):
             """A Watchdog Event Handler method."""
             self._check_calculation()
 
-    def __init__(self, stale_after, next_time, reload, cache_dir):
+    def __init__(
+            self, stale_after, next_time, reload, cache_dir, separate_files):
         _BaseCore.__init__(self, stale_after, next_time)
         self.cache = None
         self.reload = reload
@@ -94,6 +95,7 @@ class _PickleCore(_BaseCore):
         self.lock = threading.RLock()
         self.cache_fname = None
         self.cache_fpath = None
+        self.separate_files = separate_files
 
     def _cache_fname(self):
         if self.cache_fname is None:
@@ -104,7 +106,6 @@ class _PickleCore(_BaseCore):
 
     def _cache_fpath(self):
         if self.cache_fpath is None:
-            # print(EXPANDED_CACHIER_DIR)
             if not os.path.exists(self.expended_cache_dir):
                 os.makedirs(self.expended_cache_dir)
             self.cache_fpath = os.path.abspath(
@@ -133,17 +134,56 @@ class _PickleCore(_BaseCore):
                 self._reload_cache()
             return self.cache
 
-    def _save_cache(self, cache):
+    def _get_cache_by_key(self, key=None, hash=None):
+        fpath = self._cache_fpath()
+        if hash is None:
+            fpath += f'_{hashlib.sha256(pickle.dumps(key)).hexdigest()}'
+        else:
+            fpath += f'_{hash}'
+        try:
+            with portalocker.Lock(fpath, mode='rb') as cache_file:
+                try:
+                    res = pickle.load(cache_file)
+                except EOFError:
+                    res = None
+        except FileNotFoundError:
+            res = None
+        return res
+
+    def _clear_all_cache_files(self):
+        fpath = self._cache_fpath()
+        path, name = os.path.split(fpath)
+        for subpath in os.listdir(path):
+            if subpath.startswith(f'{name}_'):
+                os.remove(os.path.join(path, subpath))
+
+    def _clear_being_calculated_all_cache_files(self):
+        fpath = self._cache_fpath()
+        path, name = os.path.split(fpath)
+        for subpath in os.listdir(path):
+            if subpath.startswith(name):
+                entry = self._get_cache_by_key(hash=subpath.split('_')[-1])
+                if entry is not None:
+                    entry['being_calculated'] = False
+                    self._save_cache(entry, hash=subpath.split('_')[-1])
+
+    def _save_cache(self, cache, key=None, hash=None):
         with self.lock:
             self.cache = cache
             fpath = self._cache_fpath()
+            if key is not None:
+                fpath += f'_{hashlib.sha256(pickle.dumps(key)).hexdigest()}'
+            elif hash is not None:
+                fpath += f'_{hash}'
             with portalocker.Lock(fpath, mode='wb') as cache_file:
                 pickle.dump(cache, cache_file, protocol=4)
-            self._reload_cache()
+            if key is None:
+                self._reload_cache()
 
     def get_entry_by_key(self, key, reload=False):  # pylint: disable=W0221
         with self.lock:
-            # print('{}, {}'.format(self.reload, reload))
+            if self.separate_files:
+                return key, self._get_cache_by_key(key)
             if self.reload or reload:
                 self._reload_cache()
             return key, self._get_cache().get(key, None)
@@ -154,31 +194,53 @@ class _PickleCore(_BaseCore):
         return self.get_entry_by_key(key)
 
     def set_entry(self, key, func_res):
-        with self.lock:
-            cache = self._get_cache()
-            cache[key] = {
+        key_data = {
                 'value': func_res,
                 'time': datetime.now(),
                 'stale': False,
                 'being_calculated': False,
             }
-            self._save_cache(cache)
+        if self.separate_files:
+            self._save_cache(key_data, key)
+        else:
+            with self.lock:
+                cache = self._get_cache()
+                cache[key] = key_data
+                self._save_cache(cache)
+
+    def mark_entry_being_calculated_separate_files(self, key):
+        self._save_cache({
+                        'value': None,
+                        'time': datetime.now(),
+                        'stale': False,
+                        'being_calculated': True,
+                    }, key=key)
+
+    def mark_entry_not_calculated_separate_files(self, key):
+        _, entry = self.get_entry_by_key(key)
+        entry['being_calculated'] = False
+        self._save_cache(entry, key=key)
 
     def mark_entry_being_calculated(self, key):
-        with self.lock:
-            cache = self._get_cache()
-            try:
-                cache[key]['being_calculated'] = True
-            except KeyError:
-                cache[key] = {
-                    'value': None,
-                    'time': datetime.now(),
-                    'stale': False,
-                    'being_calculated': True,
-                }
-            self._save_cache(cache)
+        if self.separate_files:
+            self.mark_entry_being_calculated_separate_files(key)
+        else:
+            with self.lock:
+                cache = self._get_cache()
+                try:
+                    cache[key]['being_calculated'] = True
+                except KeyError:
+                    cache[key] = {
+                        'value': None,
+                        'time': datetime.now(),
+                        'stale': False,
+                        'being_calculated': True,
+                    }
+                self._save_cache(cache)
 
     def mark_entry_not_calculated(self, key):
+        if self.separate_files:
+            self.mark_entry_not_calculated_separate_files(key)
         with self.lock:
             cache = self._get_cache()
             try:
@@ -188,13 +250,18 @@ class _PickleCore(_BaseCore):
                 pass  # that's ok, we don't need an entry in that case
 
     def wait_on_entry_calc(self, key):
-        with self.lock:
-            self._reload_cache()
-            entry = self._get_cache()[key]
-            if not entry['being_calculated']:
-                return entry['value']
+        if self.separate_files:
+            entry = self._get_cache_by_key(key)
+            filename = f'{self._cache_fname()}_{hashlib.sha256(pickle.dumps(key)).hexdigest()}'
+        else:
+            with self.lock:
+                self._reload_cache()
+                entry = self._get_cache()[key]
+            filename = self._cache_fname()
+        if not entry['being_calculated']:
+            return entry['value']
         event_handler = _PickleCore.CacheChangeHandler(
-            filename=self._cache_fname(), core=self, key=key
+            filename=filename, core=self, key=key
         )
         observer = Observer()
         event_handler.inject_observer(observer)
@@ -202,19 +269,23 @@ class _PickleCore(_BaseCore):
             event_handler, path=self.expended_cache_dir, recursive=True
         )
         observer.start()
-        observer.join(timeout=1.0)
-        if observer.is_alive():
-            # print('Timedout waiting. Starting again...')
-            return self.wait_on_entry_calc(key)
+        while observer.is_alive():
+            observer.join(timeout=1.0)
         # print("Returned value: {}".format(event_handler.value))
         return event_handler.value
 
     def clear_cache(self):
-        self._save_cache({})
+        if self.separate_files:
+            self._clear_all_cache_files()
+        else:
+            self._save_cache({})
 
     def clear_being_calculated(self):
-        with self.lock:
-            cache = self._get_cache()
-            for key in cache:
-                cache[key]['being_calculated'] = False
-            self._save_cache(cache)
+        if self.separate_files:
+            self._clear_being_calculated_all_cache_files()
+        else:
+            with self.lock:
+                cache = self._get_cache()
+                for key in cache:
+                    cache[key]['being_calculated'] = False
+                self._save_cache(cache)
